#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <linux/netdevice.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#if defined(__alpha__)
#include <sys/io.h>
#endif
#include <string.h>
#include <errno.h>
#include <wait.h>
#include <signal.h>
#include <termios.h>

#include <sys/ioctl.h>

#ifndef _PC300_H
#include "pc300.h"
#endif

#include "eeprom.h"
#include "encode.h"
#include "pciutils.h"

/* PCI Definitions */
#define	PCI_VENDOR_ID_CYCLADES		0x120e
#define	PCI_DEVICE_ID_PC300_RX_2	0x0300
#define	PCI_DEVICE_ID_PC300_RX_1	0x0301
#define	PCI_DEVICE_ID_PC300_TE_2	0x0310
#define	PCI_DEVICE_ID_PC300_TE_1	0x0311
#define	PCI_CLASS_PC300			(0x07800000UL)

#ifndef min
#define	min(a,b)	(((a)<(b))?(a):(b))
#endif
#ifndef max
#define	max(a,b)	(((a)>(b))?(a):(b))
#endif

#define MEM_DEV "/dev/mem"

/****************************/
/***   Global Variables   ***/
/****************************/
static int mem_fd = -1;
static void *plxmapstart = (void *) -1;
static size_t plxmaplen = 0;

/* Communication with libpci */
static struct pci_access *pacc;		/* Device access structure */
static struct pci_filter filter;	/* Device filter */
struct pci_dev *ptpci;

struct plx_9050_eeprom
  {
	 unsigned short device_id;
	 unsigned short vendor_id;
	 unsigned short class_code1;
	 unsigned short class_code2;
	 unsigned short sub_device_id;
	 unsigned short sub_vendor_id;
	 unsigned short reserved1;
	 unsigned short int_pin;
	 unsigned short addr0_range_msw;
	 unsigned short addr0_range_lsw;
	 unsigned short addr1_range_msw;
	 unsigned short addr1_range_lsw;
	 unsigned short addr2_range_msw;
	 unsigned short addr2_range_lsw;
	 unsigned short addr3_range_msw;
	 unsigned short addr3_range_lsw;
	 unsigned short rom_range_msw;
	 unsigned short rom_range_lsw;
	 unsigned short addr0_lba_msw;  // lba=local_base_addr
	 unsigned short addr0_lba_lsw;
	 unsigned short addr1_lba_msw;
	 unsigned short addr1_lba_lsw;
	 unsigned short addr2_lba_msw;
	 unsigned short addr2_lba_lsw;
	 unsigned short addr3_lba_msw;
	 unsigned short addr3_lba_lsw;
	 unsigned short rom_lba_msw;
	 unsigned short rom_lba_lsw;
	 unsigned short addr0_brd_msw;  // brd=bus region descriptor
	 unsigned short addr0_brd_lsw;
	 unsigned short addr1_brd_msw;
	 unsigned short addr1_brd_lsw;
	 unsigned short addr2_brd_msw;
	 unsigned short addr2_brd_lsw;
	 unsigned short addr3_brd_msw;
	 unsigned short addr3_brd_lsw;
	 unsigned short rom_brd_msw;
	 unsigned short rom_brd_lsw;
	 unsigned short addr0_csbr_msw;  // csbr = chip select base and range
	 unsigned short addr0_csbr_lsw;
	 unsigned short addr1_csbr_msw;
	 unsigned short addr1_csbr_lsw;
	 unsigned short addr2_csbr_msw;
	 unsigned short addr2_csbr_lsw;
	 unsigned short addr3_csbr_msw;
	 unsigned short addr3_csbr_lsw;
	 unsigned short int_ctrl_status_msw;
	 unsigned short int_ctrl_status_lsw;
	 unsigned short eeprom_ctrl_misc_msw;
	 unsigned short eeprom_ctrl_misc_lsw;
	 char         copyright[12];	// Copyright
	 char         date[7];		// Date that the EEPROM was programmed
	 char	 eeprom_version;	// Version of the EEPROM programming
	 long 	 id0, id1;		// The 64 bit encoded ID numbers
  };

struct plx_9050_eeprom pc300_eeprom = {
	PCI_DEVICE_ID_PC300_RX_1,	// Device ID
	PCI_VENDOR_ID_CYCLADES,		// Vendor ID
	(unsigned int)((PCI_CLASS_PC300>>16) & 0x0000FFFFUL),
	(unsigned int)(PCI_CLASS_PC300 & 0x0000FFFFUL),
	PCI_DEVICE_ID_PC300_RX_1,	// Device ID
	PCI_VENDOR_ID_CYCLADES,		// Vendor ID
	0x0000U,    // Max.Latency and Min.Grant not loadable
	0x0100U,    // Interrupt Pin (Int line routing not loadable)
	0x0FFFU,    // Range for PCI to Local Address Space 0 (1Kb)
	0xFC00U,    //
	0x0FF8U,    // Range for PCI to Local Address Space 1 (512Kb)
	0x0008U,    //
	0x0000U,    // Range for PCI to Local Address Space 2
	0x0000U,    //
	0x0FFFU,    // Range for PCI to Local Address Space 3 (256b)
	0xFF00U,    //
	0x0000U,    // Range for PCI to Local Expansion ROM
	0x0000U,    //
	0x0100U,    // Local Base Address (Re-Map) for PCI to Local Address Space 0
	0x0001U,    //
	0x0000U,    // Local Base Address (Re-Map) for PCI to Local Address Space 1
	0x0001U,    //
	0x0000U,    // Local Base Address (Re-Map) for PCI to Local Address Space 2
	0x0000U,    //
	0x0000U,    // Local Base Address (Re-Map) for PCI to Local Address Space 3
	0x0001U,    //
	0x0000U,    // Local Base Address (Re-Map) for PCI to Local Expansion ROM
	0x0000U,    //
	0x0080U,    // Bus Region Descriptor for Local Address Space 0
	0x0002U,    //
	0x0080U,    // Bus Region Descriptor for Local Address Space 1
	0x0003U,    //
	0x0080U,    // Bus Region Descriptor for Local Address Space 2
	0x0002U,    //
	0x0080U,    // Bus Region Descriptor for Local Address Space 3
	0x0002U,    //
	0x0080U,    // Bus Region Descriptor for Expansion ROM Space
	0x0002U,    //
	0x0100U,    // Chip Select (CS) 0 Base and Range Register
	0x0201U,    //
	0x0004U,    // Chip Select (CS) 1 Base and Range Register
	0x0001U,    //
	0x0000U,    // Chip Select (CS) 2 Base and Range Register
	0x0000U,    //
	0x0000U,    // Chip Select (CS) 3 Base and Range Register
	0x0000U,    //
	0x0000U,    // Interrupt Control / Status Register. Enable PCI Int, Disable Local Int 2,
	0x0001U,    //      Enable Local Int 1 active low
	0x1022U,    // EEPROM Control and Misc. Control Register
	0x7092U,    //

	"C99Cyclades",	// Copyright
	"Date",			// Date that the EEPROM was programmed
	4,			// Version of the EEPROM programming
	0xFFFFFFFFUL, 0xFFFFFFFFUL	// The 64 bit encoded ID numbers
};

/****************************/
/***   General Routines   ***/
/****************************/
static void
sigint(int intno)
{
    printf ("SIGINT(%d)\n\r", intno);
    exit (-1);
}

static void
sigquit(int intno)
{
    printf ("SIGQUIT(%d)\n\r", intno);
    exit (-1);
}

int
init_pc300_eeprom (unsigned long plxbase, unsigned short type)
{
    unsigned int i;
    time_t timer;
    struct tm *tm;
    long id;
    struct plx_9050_eeprom old_eeprom, new_eeprom;

    memcpy(&new_eeprom, &pc300_eeprom, sizeof(pc300_eeprom));

    /* gets time of day */
    timer = time(NULL);
    tm = localtime(&timer);
    sprintf (new_eeprom.date, "%02d%02d%02d", 
	     (tm->tm_mon + 1), tm->tm_mday, tm->tm_year);

    /* Generate a random ID number */
    eeprom_read_block ((plxbase + 0x52), 0, &old_eeprom, sizeof(old_eeprom), 
		       1, EEPROM_PLX9050);
    if (decode(&id, old_eeprom.id0, old_eeprom.id1) == 0) {
	new_eeprom.id0 = old_eeprom.id0;
	new_eeprom.id1 = old_eeprom.id1;
    } else {
	id = (unsigned long)random();
	encode (id, &new_eeprom.id0, &new_eeprom.id1);
    }

    new_eeprom.device_id = type;
    if ((type == PCI_DEVICE_ID_PC300_TE_1) || 
	(type == PCI_DEVICE_ID_PC300_TE_2)) {
	/* Enable MMIO allocation for the FALC registers */
	new_eeprom.addr2_range_msw = 0x0FFFU;
	new_eeprom.addr2_range_lsw = 0xFC00U;
	new_eeprom.addr2_lba_msw = 0x0200U;
	new_eeprom.addr2_lba_lsw = 0x0001U;
	new_eeprom.addr2_csbr_msw = 0x0200U;
	new_eeprom.addr2_csbr_lsw = 0x0201U;

	/* Enable LINT2 for the FALC interrupt generation */
	new_eeprom.int_ctrl_status_lsw |= 0x0008U;

	/* Set USER2/CS2 as CS2 */
	new_eeprom.eeprom_ctrl_misc_lsw |= 0x0040U;

	/* Set USER0 and USER1 as inputs */
	new_eeprom.eeprom_ctrl_misc_lsw &= ~0x0012U;
    }

    /* Program the EEPROM */
    i = eeprom_write_block((plxbase + 0x52), 0, &new_eeprom, 
			   sizeof(new_eeprom), 1, EEPROM_PLX9050);

    return (i);
}

/*********************/
/***   Main Body   ***/
/*********************/
int
main (int argc, char *argv[])
{
    char dummy;
    int i;
    unsigned long plx_addr = 0L;
    unsigned char *window;
    unsigned long *plx_regP;
    int page_size = getpagesize();

    if ((mem_fd = open(MEM_DEV, O_RDWR) ) < 0){
	printf("Proto: cannot open memory via %s\n\r", MEM_DEV);
	exit (-1);
    }

    signal (SIGINT, sigint);
    signal (SIGQUIT,sigquit);

    printf("Checking PC300 boards for EEPROM updates:\n");
    pacc = pci_alloc();
    pci_filter_init(pacc, &filter);
    filter.vendor = PCI_VENDOR_ID_CYCLADES;

    pci_init(pacc);
    pci_scan_bus(pacc);
    for(i = 0, ptpci=pacc->devices; ptpci; ptpci=ptpci->next) {
	if (!pci_filter_match(&filter, ptpci)) continue;
	switch(ptpci->device_id) {
	    case PCI_DEVICE_ID_PC300_RX_1:
	    case PCI_DEVICE_ID_PC300_RX_2:
	    case PCI_DEVICE_ID_PC300_TE_1:
	    case PCI_DEVICE_ID_PC300_TE_2:
	    /* Found a valid card !! */
		i++;
		break;

	    default: /* Other Cyclades cards */
		continue;
	}

	pci_fill_info(ptpci, PCI_FILL_IDENT | PCI_FILL_BASES | PCI_FILL_IRQ);

	printf(" - PC300 #%d: ", i);
        if ((ptpci->base_addr[5] & PCI_BASE_ADDRESS_MEM_MASK) == 0) {
            /* Old EEPROM: this board should be updated */

	printf("updating EEPROM ... ");
	fflush(NULL);
	plx_addr = ptpci->base_addr[0] & PCI_BASE_ADDRESS_MEM_MASK;

	/* After initial programming, we discovered that some
	   PCI BIOS code would place these registers at addresses
	   other than page boundaries.  This code now allocates
	   a sufficiently large window that the registers can be
	   placed within the window, yet the window is on a page
	   boundary.
         */
#ifdef SYNCTEST_DEBUG
printf("synctest: mapping PLX registers.\n\r");
#endif
	plxmaplen = (sizeof(struct RUNTIME_9050) + page_size - 1)
		      & ~(page_size - 1);

#if defined(__alpha__)
	plxmapstart = mmap(0,
			   plxmaplen,
			   PROT_READ|PROT_WRITE,
			   MAP_SHARED,
			   mem_fd,
			   (off_t)((plx_addr & ~(page_size - 1)) 
				   + bus_base()))
		      + (plx_addr & (page_size - 1));
#else
	plxmapstart = mmap(0,
			   plxmaplen,
			   PROT_READ|PROT_WRITE,
			   MAP_SHARED,
			   mem_fd,
			   (off_t)(plx_addr & ~(page_size - 1))) 
		      + (plx_addr & (page_size - 1));
#endif
	window = (unsigned char *) plxmapstart;
	if((long)window <= 0){
	    printf("PC300 #%d: mmap error (plx)\n\r", i);
	    exit (-1);
	}

	plx_regP = (unsigned long *)(window);

	/* Clean up global variables */
	plxmapstart = (void *) -1;

	if (init_pc300_eeprom((unsigned long)plx_regP, ptpci->device_id) != 0)
	    printf("\n    WARNING: EEPROM programming failed. Please contact "
		   "Cyclades Technical\n"
		   "    Support at <support@cyclades.com>.\n");
	else
	    printf("OK\n");
    } else {
	printf("OK (up to date)\n");
    }
    fflush(NULL);
    }

    printf("\nThe EEPROMs in your PC300 boards have been updated.\n"
	   "You must power cycle your system before using the boards.\n");
    printf("\nPress <ENTER> to continue. ");
    dummy = getchar();

    return (0);
}/* main */

