/*****************************************************************************
* server.c	X25API: Server Appliation
*
* Author(s):	Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 1995-2000 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
//#include <linux/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <time.h>
#include <netinet/in.h>
#include <linux/if_wanpipe.h>
#include <string.h>
#include <errno.h>
#include <linux/wanpipe.h>
#include <linux/sdla_x25.h>
#include <linux/if_ether.h>

#define FALSE	0
#define TRUE	1

#define MIN_FRM_LGTH	10	
#define MAX_FRM_LGTH  	1400
#define NO_FRMS_TO_TX   10	

#define MAX_X25_ADDR_SIZE	16
#define MAX_X25_DATA_SIZE 	129
#define MAX_X25_FACL_SIZE	110

unsigned char Tx_data[MIN_FRM_LGTH + sizeof(x25api_hdr_t)];
unsigned char Rx_data[1000];

/* Prototypes */
int MakeConnection( char * );
void send_socket(int,int);
void sig_chld (int sigio);


/***************************************************************
 * Main:
 *
 *    o Make a socket connection to the driver.
 *    o Call send_socket() to transmit data
 *
 **************************************************************/   

int main(int argc, char* argv[])
{
	if (argc != 2){
		printf("Usage: ./server <card name>\n");
		exit(0);
	}

	MakeConnection(argv[argc-1]);
	return 0;
};


/***************************************************
* MakeConnection
*
*   o Create a Socket
*   o Bind a socket to a wanpipe network interface
*
* IMPORTANT:
*     No signals are allowed in this part of the code.
*     Accept ioctl call, blocks and waits on a 
*     signal which indicates that command is accepted.
*     If external signal tries to wake up the application
*     the accept ioctl call will wake up and fail the connection.
*     
*     This will be fixed in future releases.
*/         

int MakeConnection (char *i_name ) 
{
	int err = 0;
	int sock, sock1;
	struct wan_sockaddr_ll sa;
	x25api_t api_cmd;
	
	errno = 0;

	memset(&sa,0,sizeof(struct wan_sockaddr_ll));

	/* Create a new socket */
   	sock = socket( AF_WANPIPE, SOCK_RAW, 0);
   	if( sock < 0 ) {
      		perror("Socket");
      		return( FALSE );
   	} 

	/* Fill in binding information. 
	 * Before we use listen() system call
         * a socket must be binded into a virtual
         * network interface svc_listen.
         * 
         * Card name must be supplied as well. In 
         * this case the user supplies the name
         * eg: wanpipe1.
         */ 
	sa.sll_family = AF_WANPIPE;
	sa.sll_protocol = htons(0x16);
	strcpy(sa.sll_device, "svc_listen");
	strcpy(sa.sll_card, i_name);
		
	/* Bind a sock using the above address structure */
        if(bind(sock, (struct sockaddr *)&sa, sizeof(struct wan_sockaddr_ll)) == -1){
                perror("bind");
		printf("Failed to bind socket to %s interface\n",i_name);
                exit(0);
        }


	/* Put the sock into listening mode. Number 10 does not 
         * mean anything, since number of connections is depended
	 * on number of LCNs configured */

	listen(sock,10);


	for (;;){

		/* Accept call will block and wait for incoming calls.
                 * When a call comes in, a new socket will be created
                 * sock1. NOTE: that connection is not established yet.*/ 

		sock1 = accept(sock,NULL,NULL);

		if (sock1 < 0){
			printf ("Sock Accept Call FAILED BAD,no sock %i\n",sock1);
			continue;
		}	

		/* Reset the api command structure */
		memset(&api_cmd, 0, sizeof(x25api_t));

		/* The GET CALL DATA ioctl call will write in the 
                 * incoming call data into api_cmd structure. Once
                 * we have call data information, it is up to the 
                 * user to accept or clear call. */
		
		if ((err=ioctl(sock1,SIOC_WANPIPE_GET_CALL_DATA,&api_cmd)) < 0){
			printf ("Get Call Data Failed %i\n",err);
			exit(1);	
		}else{
			printf("Call data: Lcn %i, Data %s\n",
				api_cmd.hdr.lcn,api_cmd.data);
		}

		/* In this case, I accept all calls. 
                 * The ACCEPT CALL ioctl will acknoledge connecton */ 

                /* The user can send data along with the call accept
                 * indication, by submiting a filled in api_cmd 
                 * structure in the third field of the ioctl call. */

		if ((err=ioctl(sock1,SIOC_WANPIPE_ACCEPT_CALL,0)) != 0){
			printf("Accept failed %i\n",err);
			close(sock1);
			continue;
		}else{
			printf("Accept OK %i\n",api_cmd.hdr.lcn);
		}


		/* If the call is accepted, fork a child to service
                 * the call */

		if (!fork()){
			close(sock);
			send_socket(sock1,api_cmd.hdr.lcn);	
			exit(0);
		}
		close(sock1);

		/* We must check if our children have died or not
                 * otherwise our process table will be full of 
                 * zombie children */

		/* Instead of having children send a SIGCHILD signal
                 * to the server indicating their death, we poll
                 * every time a new child is created.  This is not
                 * perfect but it works :) */  

		sig_chld(0);
	}
	
	exit(0);
	return( TRUE );
}

/***************************************************
* SEND_RECEIVE_SOCKET 
*
*   o Send data throught a socket 
*   o Create a packet using x25api_t data type. 
*   o The transmit packet contains 16 bytes header 
*
*	------------------------------------------
*      |  16 bytes      |        X bytes        ...
*	------------------------------------------
* 	   Header              Data
*
*   o Data structures:
*
*     	typedef struct {
*		unsigned char  qdm	PACKED;	Q/D/M bits 
*		unsigned char  cause	PACKED;	cause field 
*		unsigned char  diagn	PACKED; diagnostics 
*		unsigned char  pktType  PACKED; 
*		unsigned short length   PACKED;
*		unsigned char  result	PACKED;
*		unsigned short lcn	PACKED;
*		char reserved[7]	PACKED;
*	}x25api_hdr_t;
*
*	typedef struct {
*		x25api_hdr_t hdr	PACKED;
*		char data[X25_MAX_DATA]	PACKED;
*	}x25api_t;
*
*
* Note: when sending DATA: driver will only look at the QDM byte of 
*                          the header.
*
* QDM: byte is a bit map of three bits:
*
* 	Q bit: Bit 0 : Qualifier bit, special kind of packet.
*                      Can be used as control packet.  
*       D bit: Bit 1 : Data acknolwedge bit. The remote will
*                      acknowledge every packet sent.
*       M bit: Bit 2 : If your packet is greater than x25 MTU,
*                      i.e 1024 bytes, than cut the packet to MTU size
*                          and set the M bit to indicate more data.                
*/

void send_socket(int sock1, int lcn)
{
	unsigned int Tx_count=0, Rx_count=0;
	unsigned short Tx_lgth;
	x25api_t* api_tx_el;
	int err=0, cnt =0, i=0;
	fd_set	writeset, readset, oobset;
	x25api_t api_cmd;


	Tx_lgth = MIN_FRM_LGTH;

	api_tx_el = (x25api_t *)Tx_data;
 
	/* Fill in the packet data */ 

	for (i=0; i<Tx_lgth ; i++){
               	api_tx_el->data[i] = (unsigned char) i;
	}


	printf ("\nCHILD STARTING TO SEND on %i\n",lcn);

	for(;;) {
		FD_ZERO(&writeset);
		FD_ZERO(&oobset);	
		FD_ZERO(&readset);

		FD_SET(sock1,&writeset);
		FD_SET(sock1,&oobset);
		FD_SET(sock1,&readset);

	
		/* The select function must be used to implement flow control.
	         * WANPIPE socket will block the user if already queued packet
                 * cannot be sent 
		 */ 

		/* Furthermore, the select function will also wait for
                 * the connection to come up.  If the connection failed
                 * we will get an OOB message indication.  Otherwise,
                 * if the connecton succeded we will be able to send or
                 * receive data. */

	        if(select(sock1 + 1, &readset, &writeset, &oobset, NULL)){

			if (FD_ISSET(sock1, &writeset)){

				err = send(sock1, Tx_data, 
						Tx_lgth+sizeof(x25api_hdr_t), 0);

				/* err contains number of bytes transmited */		
				if(err <= 0 ) {
					printf("Failed to send %i , %i\n", 
							Tx_count, err);
				}else{
					printf("Child LCN %i Packet Sent OK %i\n",
						lcn,++Tx_count);
				}
		
			}else if (FD_ISSET(sock1, &readset)){
				err = recv(sock1, Rx_data, sizeof(Rx_data), 0);

				/* err contains number of bytes transmited */		
				if(err < 0 ) {
					printf("Failed to rcv %i , %i\n", 
							Rx_count, err);
				}else{
					x25api_hdr_t *api_data = 
							(x25api_hdr_t *)Rx_data;

					printf("\tReceive OK, size: %i, qdm %x,"
						" cause %x, diagn %x, cnt: %i\n", 
						  err, api_data->qdm, 
						  api_data->cause, 
						  api_data->diagn, ++Rx_count);
				}

			}else if (FD_ISSET(sock1, &oobset)){

				printf("READING OOB MSG\n");

				/* The OOB Message will indicate that an 
				 * asynchronous event occured.  The applicaton 
			         * must stop everything and check the state of 
				 * the link.  Since link might  
                                 * have gone down */

				/* In this example I just exit. A real 
				 * application would check the state of the 
			         * sock first by reading the header information. 
				 */

				/* If we fail to read the OOB message, we can 
				 * assume that the link is down.  Thus, close 
				 * the socket and continue */

				err = recv(sock1, Rx_data, 
						sizeof(Rx_data), MSG_OOB);
		
				/* err contains number of bytes transmited */		
				if(err < 0 ) {
					printf("Failed to read OOB %i , %i\n", 
							Tx_count, err);
					close(sock1);
					return;
				}else{
					printf("Child Receive OOB MSG OK %i\n",
							++cnt);
					api_tx_el = (x25api_t *)Rx_data;

					printf("OOB Packet type %x, Cause %x,"
					    " Diagn %x, Res %x, Len %i, LCN %i\n",
						api_tx_el->hdr.pktType, 
						api_tx_el->hdr.cause,
						api_tx_el->hdr.diagn, 
						api_tx_el->hdr.result,
 						api_tx_el->hdr.length,
						api_tx_el->hdr.lcn);

					close(sock1);
					return;
				}
			}
			/* Stop after number of frames */
			if(Tx_count == NO_FRMS_TO_TX){
				/* Clear a call using CLEAR_CALL ioctl function.
				 * User can send data along with the clear 
                                 * indication by submitting x25api_t structure
                                 * in the third field of the ioctl call */
				memset(&api_cmd,0,sizeof(x25api_t));
				while(1){
					api_cmd.hdr.qdm = 0x80;
					if ((err=ioctl(sock1,SIOC_WANPIPE_CLEAR_CALL,&api_cmd)) == 0){
						printf("Clear Call OK sock %i\n",lcn);
						break;
					}
				}
				close(sock1);
				return;
			}
		}//if select
	}//For
} 


void sig_chld (int sigio)
{
	pid_t pid;
	int stat;

	while ((pid=waitpid(-1,&stat,WNOHANG)) > 0){
		printf("Child %d terminated\n",pid);
	}

	return;
}
